import os
import re
import subprocess
import time

from google.cloud import storage

from .treatment import Treatment


class Downloader:
    def __init__(self, is_test=True):
        self.client = storage.Client()
        self.is_test = is_test
        self.download_path = './Images'
        self.bucket_name = pass # Elided for privacy
        self.image_pattern = \
                re.compile(r'^(Anabolic|WildType)/(M\d)(W\d)/Wild(PTH)?\w\d\w\dN\d{1,4}Transformed\d{4}\.dcm$')
        self.fetch_blobs()


    def fetch_blobs(self):
        print('Collecting available blobs')
        bucket = self.client.get_bucket(self.bucket_name)
        self.blobs = list(bucket.list_blobs())
        print('Found ', len(self.blobs), ' blobs')


    @property
    def blobs(self):
        return self.__blobs


    @blobs.setter
    def blobs(self, value):
        self.__blobs = value


    def gs_copy(self, treatment_group):
        if treatment_group is Treatment.WILDTYPE:
            treatment = '/WildType'
        elif treatment_group is Treatment.ANABOLIC:
            treatment = '/Anabolic'
        else:
            raise ValueError('Treatment group %s not recognised' % treatment_group)

        src_parts = ['gs://', self.bucket_name, treatment]
        dst_parts = [self.download_path]

        if self.is_test:
            print('Filtering to fewer blobs for testing purposes')
            src_parts += '/M[1-2]W0'
            dst_parts += treatment

        src = ''.join(src_parts)
        dst = ''.join(dst_parts)

        if not os.path.isdir(dst):
            os.makedirs(dst, exist_ok=True)

        print(
            'Running',
            ' '.join([
                'gsutil',
                '-m',
                'cp',
                '-r',
                src,
                dst,
            ])
        )

        subprocess.call([
            'gsutil',
            '-m',
            'cp',
            '-r',
            src,
            dst,
            ])


    def download_files(self):
        print('Started downloading blobs to ', self.download_path)

        start = time.time()
        self.gs_copy(Treatment.ANABOLIC)
        self.gs_copy(Treatment.WILDTYPE)
        end = time.time()

        print(
            'Downloaded',
            len([os.path.join(dp, f) for dp, dn, fn in os.walk(self.download_path) for f in fn]),
            'files in',
            (end - start),
            'seconds',
            )
        print('Finished downloading blobs')


def filter_blob_name(filter_pattern, blob_name):
    return blob_name if filter_pattern.match(blob_name) else None


def make_full_path(image_pattern, base, blob_name):
    m = image_pattern.match(blob_name)
    (treatment_group, mouse, week, _) = m.groups()
    return os.path.join(
            base,
            treatment_group,
            ''.join([mouse, week]),
            m.string.split('/')[-1]
            )


def download_file(bucket_name, download_path, filter_pattern, blob_name):
    blob = storage.Client().bucket(bucket_name).blob(blob_name)
    full_path = make_full_path(filter_pattern, download_path, blob_name)
    root = os.path.dirname(full_path)
    if not os.path.isdir(root):
        os.makedirs(root)
    blob.download_to_filename(full_path)


if __name__ == '__main__':
    downloader = Downloader()
    downloader.download_files()
    print('Downloaded')
