import numpy as np
import os
import sys

from keras import backend as K
from tensorflow.python.lib.io import file_io

from ..differencing import Differencing
from .. import mouse_data_loader as mdl


class BaseModel:
    def __init__(self, log_dir, num_epochs, batch_size, diff_type, is_test=True):
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"
        K.set_image_data_format("channels_first")

        self.is_test = is_test
        self.all_mice = False
        self.log_dir = log_dir
        self.batch_size = batch_size
        self.num_epochs = 1 if is_test else num_epochs
        self.differencing = diff_type
        self.kernel_size = 3
        self.pool_size = 2
        self.conv_size_1 = 16
        self.conv_size_2 = 32
        self.drop_prob_conv = 0.25
        self.hidden_size_1 = 64
        self.hidden_size_2 = 32
        self.num_classes = None
        self.l2_lambda = 0.0003
        self.conv_init = "he_normal"
        self.conv_activation = "relu"
        self.conv_border = "same"
        self.dense_activation = "relu"
        self.dense_init = "he_normal"
        self.height, self.width, self.depth = mdl.get_hwd()


    def _write_array(self, filename, array_like):
        filename_full = os.path.join(self.log_dir, filename)
        with file_io.FileIO(filename_full, 'w') as f:
            np.savetxt(f, array_like, delimiter=',', fmt='%s')


    def _load_data(self):
        mdl.load_to_global_by_week(
                use_diff=self.differencing is not Differencing.NONE,
                diff=str(self.differencing),
                )

        if self.is_test:
            mdl.train_global = mdl.train_global[:1*self.batch_size]
            mdl.val_global = mdl.val_global[:1*self.batch_size]
            mdl.test_global = mdl.test_global[:1*self.batch_size]


    def _log_data_stats(self):
        self.num_classes = len(mdl.mouse_types)
        self.num_train   = float(mdl.train_global.shape[0])
        self.num_val     = float(mdl.val_global.shape[0])
        self.num_test    = float(mdl.test_global.shape[0])

        print("Percentage of training examples of each class:")
        train_labels = mdl.train_global[:, 0]
        print(sum([l == '0' for l in train_labels]) / self.num_train)
        print(sum([l == '1' for l in train_labels]) / self.num_train)
        print(sum([l == '2' for l in train_labels]) / self.num_train)
        print("\n")
        print("Percentage of validation examples of each class:")
        val_labels = mdl.val_global[:, 0]
        print(sum([l == '0' for l in val_labels]) / self.num_val)
        print(sum([l == '1' for l in val_labels]) / self.num_val)
        print(sum([l == '2' for l in val_labels]) / self.num_val)
        print("\n")
        test_labels = mdl.test_global[:, 0]
        print("Percentage of test examples of each class:")
        print(sum([l == '0' for l in test_labels]) / self.num_test)
        print(sum([l == '1' for l in test_labels]) / self.num_test)
        print(sum([l == '2' for l in test_labels]) / self.num_test)
        print("\n")
        print(self.num_train,    "training samples")
        print(self.num_val,      "validation samples")
        print(self.num_test,     "testing samples")
        print("\n")

        self._write_array('true_labels.csv', test_labels)


    def save_model(self):
        print('Saving model locally')
        self.model.save('model.h5')

        print('Copying from local to GCS')
        gs_path = os.path.join(self.log_dir, 'model.h5')
        with file_io.FileIO('model.h5', mode='rb') as src:
            with file_io.FileIO(gs_path, mode='wb') as dst:
                dst.write(src.read())
                dst.flush()
        print('Copied to GCS')


    def create_model(self):
        pass


    def run(self):
        pass
