#!/usr/bin/env/python2

################################################################################

from math import ceil

from keras import backend as K
from keras.models import Model
from keras.layers import Input, Dense, Flatten, Convolution2D, MaxPooling2D
from keras.layers import Dropout
from keras.layers.merge import Concatenate
from keras.layers.normalization import BatchNormalization
from keras.regularizers import l2

from .base_model import BaseModel
from ..differencing import Differencing
from .. import mouse_data_loader as mdl

################################################################################

class CNNTimestamps(BaseModel):
    def __init__(self, log_dir, num_epochs, batch_size, is_test=True):
        super().__init__(
                log_dir,
                num_epochs,
                batch_size,
                Differencing.NONE,
                is_test,
                )


    def create_model(self):
        input_time = Input(shape=(1,))
        input_image = Input(shape=(self.depth, self.height, self.width))

        bn_0 = BatchNormalization(axis=1)(input_image)

        # Chain 1.
        conv_1 = Convolution2D(
                self.conv_size_1,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_initializer=self.conv_init,
                kernel_regularizer=l2(self.l2_lambda),
                )(bn_0)
        conv_2 = Convolution2D(
                self.conv_size_1,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_initializer=self.conv_init,
                kernel_regularizer=l2(self.l2_lambda),
                )(conv_1)
        pool_1 = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_2)
        bn_1 = BatchNormalization()(pool_1)
        drop_1 = Dropout(self.drop_prob_conv)(bn_1)

        # Chain 2.
        conv_3 = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_initializer=self.conv_init,
                kernel_regularizer=l2(self.l2_lambda),
                )(drop_1)
        conv_4 = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_initializer=self.conv_init,
                kernel_regularizer=l2(self.l2_lambda),
                )(conv_3)
        pool_2 = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_4)
        bn_2 = BatchNormalization()(pool_2)
        drop_2 = Dropout(self.drop_prob_conv)(bn_2)

        # Chain 3.
        conv_5 = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_initializer=self.conv_init,
                kernel_regularizer=l2(self.l2_lambda),
                )(drop_2)
        conv_6 = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_initializer=self.conv_init,
                kernel_regularizer=l2(self.l2_lambda),
                )(conv_5)
        pool_3 = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_6)
        bn_3 = BatchNormalization()(pool_3)
        drop_3 = Dropout(self.drop_prob_conv)(bn_3)

        # Chain 4.
        conv_7 = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_initializer=self.conv_init,
                kernel_regularizer=l2(self.l2_lambda),
                )(drop_3)
        conv_8 = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_initializer=self.conv_init,
                kernel_regularizer=l2(self.l2_lambda),
                )(conv_7)
        pool_4 = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_8)
        bn_4 = BatchNormalization()(pool_4)
        drop_4 = Dropout(self.drop_prob_conv)(bn_4)

        # Chain 5.
        conv_9 = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_initializer=self.conv_init,
                kernel_regularizer=l2(self.l2_lambda),
                )(drop_4)
        conv_10 = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_initializer=self.conv_init,
                kernel_regularizer=l2(self.l2_lambda),
                )(conv_9)
        pool_5 = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_10)
        bn_5 = BatchNormalization()(pool_5)

        flat = Flatten()(bn_5)
        concat = Concatenate(axis=1)([flat, input_time])

        hidden_1 = Dense(
                self.hidden_size_1,
                activation=self.dense_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.dense_init,
                )(concat)

        bn_6 = BatchNormalization()(hidden_1)

        hidden_2 = Dense(
                self.hidden_size_2,
                activation=self.dense_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.dense_init,
                )(bn_6)

        bn_7 = BatchNormalization()(hidden_2)

        out = Dense(
                self.num_classes,
                activation="softmax",
                kernel_initializer="glorot_uniform",
                )(bn_7)

        self.model = Model(inputs=[input_time, input_image], outputs=[out])


    def run(self):
        self._load_data()
        self._log_data_stats()
        self.create_model()

        self.model.compile(
                loss='categorical_crossentropy',
                optimizer='adam',
                metrics=['accuracy'],
                )

        print(self.model.summary())

        print('Running training')

        self.model.fit_generator(
                mdl.generate_mouse_data(
                    which="train",
                    resize=True,
                    t=True
                    ),
                steps_per_epoch=ceil(self.num_train / float(self.batch_size)),
                epochs=self.num_epochs,
                validation_data=mdl.generate_mouse_data(
                    which="val",
                    resize=True,
                    t=True
                    ),
                validation_steps=(self.num_val / self.batch_size),
                verbose=2,
                )

        self.save_model()

        print('Running prediction')

        predicted_labels = self.model.predict_generator(
                mdl.generate_mouse_data(
                    which='test',
                    resize=True,
                    t=True,
                    ),
                steps=(self.num_test / self.batch_size),
                )

        print('Writing out labels')

        self._write_array(
                'predicted_labels.csv',
                K.argmax(predicted_labels).eval(session=K.get_session())
                )

################################################################################
