#!/usr/bin/env python3

################################################################################

from math import ceil

from keras import backend as K
from keras.models import Model
from keras.layers import Dense, Dropout, Flatten
from keras.layers import Convolution2D, MaxPooling2D, Input
from keras.layers.merge import Concatenate
from keras.layers.normalization import BatchNormalization
from keras.regularizers import l2

from .base_model import BaseModel
from ..differencing import Differencing
from .. import mouse_data_loader as mdl

################################################################################

class CNNAbsoluteDifferencing(BaseModel):
    def __init__(self, log_dir, num_epochs, batch_size, is_test=True):
        super().__init__(
                log_dir,
                num_epochs,
                batch_size,
                Differencing.ABSOLUTE,
                is_test,
                )
        self.conv_size_1 = 8
        self.conv_size_2 = 16
        self.conv_size_3 = 32


    def create_model(self):
        input_image  = Input(shape=(self.depth, self.height, self.width))
        input_diff = Input(shape=(self.depth, self.height, self.width))

        bn_0_i = BatchNormalization(axis=1)(input_image)
        bn_0_d = BatchNormalization(axis=1)(input_diff)

        # Chain 1.
        conv_1_i = Convolution2D(
                self.conv_size_1,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(bn_0_i)
        conv_1_d = Convolution2D(
                self.conv_size_1,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(bn_0_d)

        conv_2_i = Convolution2D(
                self.conv_size_1,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(conv_1_i)
        conv_2_d = Convolution2D(
                self.conv_size_1,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(conv_1_d)

        pool_1_i = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_2_i)
        pool_1_d = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_2_d)

        bn_1_i = BatchNormalization(axis=1)(pool_1_i)
        bn_1_d = BatchNormalization(axis=1)(pool_1_d)

        drop_1_i = Dropout(self.drop_prob_conv)(bn_1_i)
        drop_1_d = Dropout(self.drop_prob_conv)(bn_1_d)

        # Cross-connections.
        i_to_d_1 = Convolution2D(
                self.conv_size_1,
                (1, 1),
                padding=self.conv_border,
                activation="relu",
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(drop_1_i)
        d_to_i_1 = Convolution2D(
                self.conv_size_1,
                (1, 1),
                padding=self.conv_border,
                activation="relu",
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(drop_1_d)

        i_map_1 = Concatenate(axis=1)([drop_1_i, d_to_i_1])
        d_map_1 = Concatenate(axis=1)([drop_1_d, i_to_d_1])

        # Chain 2.
        conv_3_i = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(i_map_1)
        conv_3_d = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(d_map_1)

        conv_4_i = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(conv_3_i)
        conv_4_d = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(conv_3_d)

        pool_2_i = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_4_i)
        pool_2_d = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_4_d)

        bn_2_i = BatchNormalization(axis=1)(pool_2_i)
        bn_2_d = BatchNormalization(axis=1)(pool_2_d)

        drop_2_i = Dropout(self.drop_prob_conv)(bn_2_i)
        drop_2_d = Dropout(self.drop_prob_conv)(bn_2_d)

        # Cross-connections.
        i_to_d_2 = Convolution2D(
                self.conv_size_1,
                (1, 1),
                padding=self.conv_border,
                activation="relu",
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(drop_2_i)
        d_to_i_2 = Convolution2D(
                self.conv_size_1,
                (1, 1),
                padding=self.conv_border,
                activation="relu",
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(drop_2_d)

        i_map_2 = Concatenate(axis=1)([drop_2_i, d_to_i_2])
        d_map_2 = Concatenate(axis=1)([drop_2_d, i_to_d_2])

        # Chain 3.
        conv_5_i = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(i_map_2)
        conv_5_d = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(d_map_2)

        conv_6_i = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(conv_5_i)
        conv_6_d = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(conv_5_d)

        pool_3_i = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_6_i)
        pool_3_d = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_6_d)

        bn_3_i = BatchNormalization(axis=1)(pool_3_i)
        bn_3_d = BatchNormalization(axis=1)(pool_3_d)

        drop_3_i = Dropout(self.drop_prob_conv)(bn_3_i)
        drop_3_d = Dropout(self.drop_prob_conv)(bn_3_d)

        # Cross-connections.
        i_to_d_3 = Convolution2D(
                self.conv_size_1,
                (1, 1),
                padding=self.conv_border,
                activation="relu",
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(drop_3_i)
        d_to_i_3 = Convolution2D(
                self.conv_size_1,
                (1, 1),
                padding=self.conv_border,
                activation="relu",
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(drop_3_d)

        i_map_3 = Concatenate(axis=1)([drop_3_i, d_to_i_3])
        d_map_3 = Concatenate(axis=1)([drop_3_d, i_to_d_3])

        # Chain 4.
        conv_7_i = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(i_map_3)
        conv_7_d = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(d_map_3)

        conv_8_i = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(conv_7_i)
        conv_8_d = Convolution2D(
                self.conv_size_2,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(conv_7_d)

        pool_4_i = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_8_i)
        pool_4_d = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_8_d)

        bn_4_i = BatchNormalization(axis=1)(pool_4_i)
        bn_4_d = BatchNormalization(axis=1)(pool_4_d)

        drop_4_i = Dropout(self.drop_prob_conv)(bn_4_i)
        drop_4_d = Dropout(self.drop_prob_conv)(bn_4_d)

        # Cross-connections.
        i_to_d_4 = Convolution2D(
                self.conv_size_1,
                (1, 1),
                padding=self.conv_border,
                activation="relu",
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(drop_4_i)
        d_to_i_4 = Convolution2D(
                self.conv_size_1,
                (1, 1),
                padding=self.conv_border,
                activation="relu",
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(drop_4_d)

        i_map_4 = Concatenate(axis=1)([drop_4_i, d_to_i_4])
        d_map_4 = Concatenate(axis=1)([drop_4_d, i_to_d_4])

        # Chain 5.
        conv_9_i = Convolution2D(
                self.conv_size_3,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(i_map_4)
        conv_9_d = Convolution2D(
                self.conv_size_3,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(d_map_4)

        conv_10_i = Convolution2D(
                self.conv_size_3,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(conv_9_i)
        conv_10_d = Convolution2D(
                self.conv_size_3,
                (self.kernel_size, self.kernel_size),
                padding=self.conv_border,
                activation=self.conv_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.conv_init,
                )(conv_9_d)

        pool_5_i = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_10_i)
        pool_5_d = MaxPooling2D(
                pool_size=(self.pool_size, self.pool_size),
                strides=(self.pool_size, self.pool_size),
                padding=self.conv_border,
                )(conv_10_d)

        bn_5_i = BatchNormalization(axis=1)(pool_5_i)
        bn_5_d = BatchNormalization(axis=1)(pool_5_d)

        flat_i = Flatten()(bn_5_i)
        flat_d = Flatten()(bn_5_d)
        concat = Concatenate(axis=1)([flat_i, flat_d])

        hidden_1 = Dense(
                self.hidden_size_1,
                activation=self.dense_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.dense_init,
                )(concat)

        bn_6 = BatchNormalization()(hidden_1)

        hidden_2 = Dense(
                self.hidden_size_2,
                activation=self.dense_activation,
                kernel_regularizer=l2(self.l2_lambda),
                kernel_initializer=self.dense_init,
                )(bn_6)

        bn_7 = BatchNormalization()(hidden_2)

        out = Dense(self.num_classes,
                    activation="softmax",
                    init="glorot_uniform",
                    )(bn_7)

        self.model = Model(
                inputs=[input_image, input_diff],
                outputs=[out],
                )


    def run(self):
        self._load_data()
        self._log_data_stats()
        self.create_model()

        self.model.compile(
                loss='categorical_crossentropy',
                optimizer='adam',
                metrics=['categorical_accuracy']
                )

        print(self.model.summary())

        print('Running training')

        self.model.fit_generator(
                mdl.generate_diff_mouse_data(
                    which="train",
                    resize=True,
                    t=False,
                    diff="abs"
                    ),
                steps_per_epoch=ceil(self.num_train / self.batch_size),
                epochs=self.num_epochs,
                validation_data=mdl.generate_diff_mouse_data(
                    which="val",
                    resize=True,
                    t=False,
                    diff="abs"
                    ),
                validation_steps=(self.num_val / self.batch_size),
                verbose=2
                )

        self.save_model()

        print('Running prediction')

        predicted_labels = self.model.predict_generator(
                mdl.generate_diff_mouse_data(
                    which="test",
                    resize=True,
                    t=False,
                    diff="abs",
                    ),
                steps=(self.num_test / self.batch_size)
                )

        print('Writing out labels')

        self._write_array(
                'predicted_labels.csv',
                K.argmax(predicted_labels).eval(session=K.get_session())
                )

################################################################################
